﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Reflection;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Markup;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using System.Xml;

namespace TextOutlineGenerator
{
    public partial class MainWindow : Window
    {
        Typeface typeface;
        FlowDirection direction;
        CultureInfo culture;
        double fontSize;
        FormattedText formattedText;
        string Text;
        GeometryType geometryType;
        bool separateIntoLetters, separateIntoComponents, formatOutput = true, silverlight;

        [STAThread]
        public static void Main()
        {
            Application app = new Application();
            app.Run(new MainWindow());
        }

        public MainWindow()
        {
            InitializeComponent();

            comboFamily.ItemsSource = Fonts.SystemFontFamilies;
            comboFamily.SelectedItem = FontFamily;

            FillWithStaticMembers(comboStyle, typeof(FontStyles), FontStyle);
            FillWithStaticMembers(comboWeight, typeof(FontWeights), FontWeight);
            FillWithStaticMembers(comboStretch, typeof(FontStretches), FontStretch);

            comboDirection.ItemsSource = Enum.GetValues(typeof(FlowDirection));
            comboDirection.SelectedItem = FlowDirection;

            comboCulture.ItemsSource = CultureInfo.GetCultures(CultureTypes.AllCultures);
            comboCulture.SelectedItem = CultureInfo.CurrentCulture;
            
            txtboxSize.Text = (0.75 * FontSize).ToString();
            txtboxText.Text = "HELLO WORLD";

            geometryRadioButton.IsChecked = true;
            formatOuputCheckBox.IsChecked = true;
        }

        void FillWithStaticMembers(ComboBox comboBox, Type typeClass, object value)
        {
            PropertyInfo[] properties = typeClass.GetProperties(BindingFlags.Public | 
                                                                BindingFlags.Static);

            foreach (PropertyInfo property in properties)
                comboBox.Items.Add(property.GetValue(null, null));

            comboBox.SelectedItem = value;
        }

        void OnTypefaceComboSelectionChanged(object sender, SelectionChangedEventArgs args)
        {
            formattedText = null;
            typeface = null;
            formattedTextBlock.FormattedText = formattedText;
            txtblkError.Text = "";
            txtboxResult.Text = "";

            if (comboFamily == null || comboStyle == null || comboWeight == null || comboStretch == null)
            {
                txtblkError.Text = "Typeface-related ComboBox object is null";
                return;
            }

            if (comboFamily.SelectedItem == null)
            {
                txtblkError.Text = "Family ComboBox has no selected item";
                return;
            }

            if (comboStyle.SelectedItem == null)
            {
                txtblkError.Text = "Style ComboBox has no selected item";
                return;
            }

            if (comboWeight.SelectedItem == null)
            {
                txtblkError.Text = "Weight ComboBox has no selected item";
                return;
            }

            if (comboStretch.SelectedItem == null)
            {
                txtblkError.Text = "Stretch ComboBox has no selected item";
                return;
            }

            try
            {
                typeface = new Typeface((FontFamily)comboFamily.SelectedItem,
                                        (FontStyle)comboStyle.SelectedItem,
                                        (FontWeight)comboWeight.SelectedItem,
                                        (FontStretch)comboStretch.SelectedItem);
            }
            catch (Exception exc)
            {
                txtblkError.Text = "Could not create Typeface: " + exc.Message;
                return;
            }

            if (typeface == null)
            {
                txtblkError.Text = "Typeface is null for no apparent reason";
                return;
            }

            DisplayText();
        }

        void OnOtherComboSelectionChanged(object sender, SelectionChangedEventArgs args)
        {
            formattedText = null;
            formattedTextBlock.FormattedText = formattedText;
            txtblkError.Text = "";
            txtboxResult.Text = "";

            if (comboCulture == null || comboDirection == null)
            {
                txtblkError.Text = "FormattedText-related ComboBox object is null";
                return;
            }

            if (comboDirection.SelectedItem == null)
            {
                txtblkError.Text = "Direction ComboBox has no selected item";
                return;
            }

            direction = (FlowDirection)comboDirection.SelectedItem;

            if (comboCulture.SelectedItem == null)
            {
                txtblkError.Text = "Culture ComboBox has no selected item";
                return;
            }

            culture = (CultureInfo)comboCulture.SelectedItem;

            DisplayText();
        }

        void OnSizeTextChanged(object sender, TextChangedEventArgs args)
        {
            formattedText = null;
            formattedTextBlock.FormattedText = formattedText;
            txtblkError.Text = "";
            txtboxResult.Text = "";
            fontSize = 0;

            if (!Double.TryParse(txtboxSize.Text, out fontSize))
            {
                txtblkError.Text = "Size is not a valid value";
                return;
            }

            fontSize /= 0.75;

            DisplayText();
        }

        void OnTextTextChanged(object sender, TextChangedEventArgs args)
        {
            formattedText = null;
            formattedTextBlock.FormattedText = formattedText;
            txtblkError.Text = "";
            txtboxResult.Text = "";

            Text = txtboxText.Text;

            DisplayText();
        }

        void OnGeometryTypeRadioButtonChecked(object sender, RoutedEventArgs args)
        {
            RadioButton radio = sender as RadioButton;
            geometryType = (GeometryType)radio.Tag;
            GenerateOutlineGeometry();
        }

        void OnSeparateLettersCheckBoxChecked(object sender, RoutedEventArgs args)
        {
            separateIntoLetters = (bool)(sender as CheckBox).IsChecked;
            GenerateOutlineGeometry();
        }

        void OnSeparateComponentsCheckBoxChecked(object sender, RoutedEventArgs args)
        {
            separateIntoComponents = (bool)(sender as CheckBox).IsChecked;
            GenerateOutlineGeometry();
        }

        void OnFormatOutputCheckBoxChecked(object sender, RoutedEventArgs args)
        {
            formatOutput = (bool)(sender as CheckBox).IsChecked;
            txtboxResult.TextWrapping = formatOutput ? TextWrapping.NoWrap : TextWrapping.Wrap;
            GenerateOutlineGeometry();
        }

        void OnSilverlightCheckBoxChecked(object sender, RoutedEventArgs args)
        {
            silverlight = (bool)(sender as CheckBox).IsChecked;
            GenerateOutlineGeometry();
        }

        void DisplayText()
        {
            formattedText = null;
            if (formattedTextBlock != null)
                formattedTextBlock.FormattedText = formattedText;

            txtblkError.Text = "";
            txtboxResult.Text = "";

            if (typeface == null || fontSize == 0 || String.IsNullOrEmpty(Text))
            {
                return;
            }

            try
            {
                formattedText = new FormattedText(Text, culture, direction, typeface, fontSize, Brushes.Black);
            }
            catch (Exception exc)
            {
                txtblkError.Text = "FormattedText could not be created: " + exc.Message;
            }

            if (formattedText == null)
            {
                txtblkError.Text = "FormattedText is null for no apparent reason";
                return;
            }

            formattedTextBlock.FormattedText = formattedText;
            GenerateOutlineGeometry();
        }

        void GenerateOutlineGeometry()
        {
            txtboxResult.Text = "";
            txtboxTextSize.Text = "";

            if (formattedText == null)
                return;

            if (separateIntoLetters)
            {
                double x = 0;

                foreach (Char ch in formattedText.Text)
                {
                    FormattedText letterText = new FormattedText(ch.ToString(), culture, direction, typeface, fontSize, Brushes.Black);
                    BuildGeometry(letterText, new Point(x, 0));
                    x += letterText.WidthIncludingTrailingWhitespace;
                    txtboxResult.Text += "\r\n\r\n";
                    txtboxTextSize.Text += new Size(letterText.WidthIncludingTrailingWhitespace, letterText.Height).ToString() + " ";
                }
            }
            else
            {
                BuildGeometry(formattedText, new Point(0, 0));
                txtboxTextSize.Text = new Size(formattedText.WidthIncludingTrailingWhitespace, formattedText.Height).ToString();
            }
        }

        void BuildGeometry(FormattedText formattedText, Point point)
        {
            Geometry geometry = formattedText.BuildGeometry(point);

            if (geometryType == GeometryType.PathGeometry)
                geometry = PathGeometry.CreateFromGeometry(geometry);

            else if (geometryType == GeometryType.FlattenedPathGeometry)
                geometry = geometry.GetFlattenedPathGeometry();

            if (geometry == null)
                return;

            if (separateIntoComponents && geometry is GeometryGroup)
            {
                GeometryGroup geoGroup = geometry as GeometryGroup;

                while (geoGroup.Children.Count == 1 && geoGroup.Children[0] is GeometryGroup)
                    geoGroup = geoGroup.Children[0] as GeometryGroup;

                foreach (Geometry geo in geoGroup.Children)
                {
                    string str = GetXaml(geo);
                    txtboxResult.Text += str;
                    txtboxResult.Text += "\r\n\r\n";
                }
            }

            else if (separateIntoComponents && geometry is PathGeometry)
            {
                PathGeometry pathGeometry = geometry as PathGeometry;

                if (silverlight)
                {
                    pathGeometry = pathGeometry.Clone();
                }

                foreach (PathFigure pathFigure in pathGeometry.Figures)
                {
                    if (silverlight)
                    {
                        foreach (PathSegment pathSegment in pathFigure.Segments)
                        {
                            pathSegment.ClearValue(PathSegment.IsSmoothJoinProperty);
                            pathSegment.ClearValue(PathSegment.IsStrokedProperty);
                        }
                    }

                    string str = GetXaml(pathFigure);
                    txtboxResult.Text += str;
                    txtboxResult.Text += "\r\n\r\n";
                }
            }

            else
            {
                string str = GetXaml(geometry);
                txtboxResult.Text += str;
            }
        }

        string GetXaml(object obj)
        {
            if (formatOutput)
            {
                StringBuilder builder = new StringBuilder();
                XmlWriterSettings settings = new XmlWriterSettings()
                {
                    OmitXmlDeclaration = true,
                    Indent = true,
                    IndentChars = "    ",
                    NewLineOnAttributes = true
                };

                XmlWriter writer = XmlWriter.Create(builder, settings);
                XamlWriter.Save(obj, writer);
                return builder.ToString();
            }
            else
            {
                return XamlWriter.Save(obj);
            }
        }
    }
}
